@@ -18,7 +18,7 @@ import (
 	"reflect"
 	"testing"
 
-	pb "github.com/coreos/etcd/raft/raftpb"
+	pb "go.etcd.io/etcd/raft/raftpb"
 )
 
 func TestFindConflict(t *testing.T) {
@@ -146,12 +146,14 @@ func TestAppend(t *testing.T) {
 
 // TestLogMaybeAppend ensures:
 // If the given (index, term) matches with the existing log:
-// 	1. If an existing entry conflicts with a new one (same index
-// 	but different terms), delete the existing entry and all that
-// 	follow it
-// 	2.Append any new entries not already in the log
+//  1. If an existing entry conflicts with a new one (same index
+//     but different terms), delete the existing entry and all that
+//     follow it
+//     2.Append any new entries not already in the log
+//
 // If the given (index, term) does not match with the existing log:
-// 	return false
+//
+//	return false
 func TestLogMaybeAppend(t *testing.T) {
 	previousEnts := []pb.Entry{{Index: 1, Term: 1}, {Index: 2, Term: 2}, {Index: 3, Term: 3}}
 	lastindex := uint64(3)
@@ -265,7 +267,7 @@ func TestLogMaybeAppend(t *testing.T) {
 					t.Fatalf("unexpected error %v", err)
 				}
 				if !reflect.DeepEqual(tt.ents, gents) {
-					t.Errorf("%d: appended entries = %v, want %v", i, gents, tt.ents)
+					t.Errorf("#%d: appended entries = %v, want %v", i, gents, tt.ents)
 				}
 			}
 		}()
@@ -282,11 +284,11 @@ func TestCompactionSideEffects(t *testing.T) {
 	lastTerm := lastIndex
 	storage := NewMemoryStorage()
 	for i = 1; i <= unstableIndex; i++ {
-		storage.Append([]pb.Entry{{Term: uint64(i), Index: uint64(i)}})
+		storage.Append([]pb.Entry{{Term: i, Index: i}})
 	}
 	raftLog := newLog(storage, raftLogger)
 	for i = unstableIndex; i < lastIndex; i++ {
-		raftLog.append(pb.Entry{Term: uint64(i + 1), Index: uint64(i + 1)})
+		raftLog.append(pb.Entry{Term: i + 1, Index: i + 1})
 	}
 
 	ok := raftLog.maybeCommit(lastIndex, lastTerm)
@@ -426,7 +428,7 @@ func TestUnstableEnts(t *testing.T) {
 
 		ents := raftLog.unstableEntries()
 		if l := len(ents); l > 0 {
-			raftLog.stableTo(ents[l-1].Index, ents[l-i].Term)
+			raftLog.stableTo(ents[l-1].Index, ents[l-1].Term)
 		}
 		if !reflect.DeepEqual(ents, tt.wents) {
 			t.Errorf("#%d: unstableEnts = %+v, want %+v", i, ents, tt.wents)
@@ -528,7 +530,7 @@ func TestStableToWithSnap(t *testing.T) {
 	}
 }
 
-//TestCompaction ensures that the number of log entries is correct after compactions.
+// TestCompaction ensures that the number of log entries is correct after compactions.
 func TestCompaction(t *testing.T) {
 	tests := []struct {
 		lastIndex uint64
@@ -671,13 +673,13 @@ func TestIsOutOfBounds(t *testing.T) {
 			}()
 			err := l.mustCheckOutOfBounds(tt.lo, tt.hi)
 			if tt.wpanic {
-				t.Errorf("%d: panic = %v, want %v", i, false, true)
+				t.Errorf("#%d: panic = %v, want %v", i, false, true)
 			}
 			if tt.wErrCompacted && err != ErrCompacted {
-				t.Errorf("%d: err = %v, want %v", i, err, ErrCompacted)
+				t.Errorf("#%d: err = %v, want %v", i, err, ErrCompacted)
 			}
 			if !tt.wErrCompacted && err != nil {
-				t.Errorf("%d: unexpected err %v", i, err)
+				t.Errorf("#%d: unexpected err %v", i, err)
 			}
 		}()
 	}
